'     PROGRAM CLUSBAS
'-----------------------------------------------------------------------
'  This is a simple program well suited for cluster analysis of
'  variables.  The clustering method used is average-linkage
'  hierarchical cluster analysis.  There are few options, but that
'  provides the advantage of easy use.  All commands are supplied to
'  the program interactively, in response to prompting questions.
'
'  Reference:  Anderberg, M. R. (1973).  Cluster analysis for
'              applications.  New York:  Academic press.
'
'  Almost any text on cluster analysis should have a good
'  description of the average-linkage hierarchical clustering
'  algorithm.   The algorithm begins with an initial similarity
'  or dissimilarity matrix between pairs of objects.  The
'  algorithm proceeds in an iterative way.  At each iteration
'  the two most similar (we assume similarities for explanation)
'  objects are combined into one group.  At each successive
'  iteration, the two most similar objects or groups of objects are
'  merged.  Similarity between groups is defined as the average
'  similarity between objects in one group with objects in the other.
'
'     INPUT:   A correlation matrix (or some other similarity or
'              dissimilarity matrix) in a file named MATRIX.DAT
'              This must contain all the elements of a full
'              (n x n), symmetrical matrix.  Any format is
'              allowable, as long as numbers are separated by
'              blanks.
'
'     OUTPUT:  Output consists of a cluster history and a tree
'              diagram (dendogram).  The cluster history
'              indicates, for each iteration, the objects
'              or clusters merged, and the average pairwise
'              similarity or dissimilarity in the resulting
'              cluster.
'
'     Two utility files are created by the program:  SORTED.DAT
'     and CLUM.DAT.  These can be deleted after running the program.
'     If they are present from a previous run they will be re-written
'     and so will not affect the results of a current run.
'-----------------------------------------------------------------------
'  Notes:
'
'  1.  Compatible with Microsoft QuickBASIC, versions 4.0 or
'      above.
'
'  2.  This comes from a pre FORTRAN-77 program--please excuse
'      lack of block command structure.
'
'  3.  The original version was also written as several separate
'      programs--which, for example, explains why files are used
'      to communicate data between subroutines.
'
'  4.  QuickBASIC apparently lists the subroutines in alphabetic
'      order, regardless of how they are presented in the original
'      code.  Please excuse this inconvenience.
'
'  5.  The dendogram is formatted for 132-column line printers.
'      If your printer can print longer lines (e.g., with compressed
'      type you, can reset the ISCL values in subroutine TREE; see
'      comments in there flagged "**SCALING**".
'
'  6.  The program may be freely copied and distributed.
'
'  Author:     John Uebersax
'  Written:    1978
'
'  Revised:    Change(s):
'  ---------   ----------------------------------------------------------
'  17 mar 93   Translated from FORTRAN
'  22 mar 93   More comments added
'  10 jun 06   1 - change output file to output.txt
'-----------------------------------------------------------------------
      DECLARE SUB RMAKE (X$, LIMIT!, ISTART!)
      DECLARE SUB TREE ()
      DECLARE SUB CONV (N1!, N2!, A$)
      DECLARE SUB PRETRE ()
      DECLARE SUB CLUSV1 ()
      DECLARE SUB SCAN ()
      DECLARE SUB BSCAN ()
      DECLARE SUB ARRANGE ()
'-----------------------------------------------------------------------
'  Driver program
'
      DIM X(100, 100)   ' Element X(i, j) is the similarity or
                        ' dissimilarity between object or group i
                        ' and object or group j
      DIM NIN(100)      ' NIN(i) is the number of objects in
                        ' cluster I
      DIM NVAR(100)     ' Used to identify clusters
      DIM ROW$(200)     ' For output--a line of the dendogram
 
      COMMON SHARED X(), NIN(), NVAR(), K, L, M, RX, ROW$()
 
      CALL CLUSV1       ' CLUSV1 routine
      CALL PRETRE       ' PRETRE routine
      CALL TREE         ' TREE routine
      END
 
      SUB ARRANGE
'
' This subroutine updates the similarity or dissimilarity matrix.
' If two objects/groups K and L are merged, it calculates the
' similarity or dissimilarity of the new group with all other objects
' or groups.  It does this by averaging the elements in row K of
' X() with those in row L, and similarly for columns K and L.
' The new elements are put in row K and column L (K < L).  Row K
' and column L are deleted.  Columns and rows greater than L are
' shifted up one column or row to fill in the gap.  The resulting
' matrix X() thus has one less column and row then at the beginning
' of the subroutine.
'
      MN = M - 1
      SAV = X(K, L)
      SAV2 = X(K, K)
'
' Calculate similarity or dissimilarity of group formed by merging I
' and J to all other groups by averaging the similarities or
' dissimilarities of I and J with other groups
'
      FOR I = 1 TO M
         X(I, K) = (X(I, K) * NIN(K) + X(I, L) * NIN(L)) / (NIN(K) + NIN(L))
         X(K, I) = X(I, K)
      NEXT I
      X(K, K) = SAV2 * NIN(K) * (NIN(K) - 1) + X(L, L) * NIN(L) * (NIN(L) - 1)
      X(K, K) = X(K, K) + SAV * 2 * NIN(K) * NIN(L)
      X(K, K) = X(K, K) / ((NIN(K) + NIN(L)) * (NIN(K) + NIN(L) - 1))
      IF (L = M) THEN 60
      FOR I = 1 TO M
'
' Shift columns after J up one place
'
         FOR J = L TO MN
            X(I, J) = X(I, J + 1)
         NEXT J
      NEXT I
      FOR I = L TO MN
'
' Shift rows after J up one place
'
         FOR J = 1 TO M
            X(I, J) = X(I + 1, J)
         NEXT J
      NEXT I
 
' Update number of objects in each cluster
'
      NIN(K) = NIN(K) + NIN(L)
      FOR I = L TO MN
         NIN(I) = NIN(I + 1)
      NEXT I
      GOTO 70
60    NIN(K) = NIN(K) + NIN(L)
70    END SUB
 
      SUB BSCAN
'
' This subroutine looks for the minimum dissimilarity.  It finds
' element (K, L), where K and L are the most dissimilar objects
' or groups.
'
      N = 1
      RRRMIN = 1000000!
      MN = M - 1
      FOR I = 1 TO MN
         N = N + 1
         FOR J = N TO M
            IF ((RRRMIN - X(I, J) < 0!)) THEN 1010
            IF ((RRRMIN - X(I, J) = 0!)) THEN 1005
            IF ((RRRMIN - X(I, J) > 0!)) THEN 1005
1005        K = I
            L = J
            RRRMIN = X(I, J)
1010     NEXT J
      NEXT I
      RX = RRRMIN
      END SUB
 
     SUB CLUSV1
'
' This routine does the cluster analysis, taking data from file
' <MATRIX.DAT>.  Parameters controlling the analysis are obtained
' interactively.  The results are stored in file <OUTPUT.TXT>, and
' also in <CLUM.DAT>, which is used as an input file for the routine
' PRETRE.
'
      DIS$ = "DIS"
 
      OPEN "CLUM.DAT" FOR OUTPUT AS #2
      OPEN "MATRIX.DAT" FOR INPUT AS #3
      OPEN "OUTPUT.TXT" FOR OUTPUT AS #7
'
' Read clustering parameters
'
      CLS
      PRINT "Interactive cluster analysis"
      PRINT " "
      INPUT "What is the problem title? ", TITLE$
      PRINT #7, TITLE$
      PRINT #2, TITLE$
 
      PRINT ""
      INPUT "How many things are to be clustered? ", M
      PRINT #2, M
 
      LIMIT = M - 1
 
      PRINT ""
      INPUT "SIM or DIS data? ", CRIT$
      CRIT$ = UCASE$(CRIT$)
      PRINT #2, CRIT$
'
' Initialize X, NIN, and NVAR
'
      FOR I = 1 TO M
         FOR J = 1 TO M
            INPUT #3, X(I, J)
         NEXT J
      NEXT I
 
      FOR I = 1 TO M
         NVAR(I) = I
         NIN(I) = 1
      NEXT I
'
' cluster analysis
'
      ITR = 0
300   ITR = ITR + 1
'
' determine groups to be merged this iteration
'
      IF (CRIT$ = DIS$) THEN GOTO 311
      CALL SCAN
      GOTO 312
311   CALL BSCAN
'
' update inter-group similarity/dissimilarity matrix taking into
' account the merger occurring this iteration
'
312   CALL ARRANGE
 
'
' OUTPUT
'
      IF (CRIT$ = "DIS") THEN 400
      PRINT #7, ""
      PRINT #7, USING " GROUP ### IS JOINED BY GROUP ###. "; NVAR(K), NVAR(L);
      PRINT #7, USING " N IS ###  ITER = ###  "; NIN(K), ITR;
      PRINT #7, USING " SIM = ######.###"; RX
      GOTO 410
400   PRINT #7, ""
      PRINT #7, USING " GROUP ### IS JOINED BY GROUP ###. "; NVAR(K), NVAR(L);
      PRINT #7, USING " N IS ###  ITER = ###  "; NIN(K), ITR;
      PRINT #7, USING " DIS = ######.###"; RX
410   PRINT #2, USING " ### ### ####.###"; NVAR(K), NVAR(L), RX
      IF (L = M) THEN 600
      MN = M - 1
      FOR I = L TO MN
      NVAR(I) = NVAR(I + 1)
      NEXT I
600   M = M - 1
      IF (ITR < LIMIT) THEN 300
999   PRINT ""
      PRINT "OUTPUT STORED ON FILE <OUTPUT.TXT>"
      CLOSE 2
      CLOSE 3
      END SUB
 
      SUB CONV (N1, N2, A$)
'
' This subroutine is used by SUB TREE.  It simply converts object
' ID numbers to string format for labeling the branches of the
' dendogram.
'
' N1 = THREE DIGIT INTEGER
' N2 = DECIMAL PLACE
 
      SWITCH = 0!
      R = N1 / 1000!
      K = 4 - N2
      FOR I = 1 TO K
         X = R * 10!
         J = INT(X)
         R = X - J + .0001
         IF (J > 0) THEN SWITCH = 1!
         IF (J = 0) THEN J = 10
         IF (J = 10 AND SWITCH = 0! AND I <> 3) THEN J = 11
         IF (J < 10) THEN A$ = RIGHT$(STR$(J), 1)
         IF (J = 10) THEN A$ = "0"
         IF (J = 11) THEN A$ = " "
      NEXT I
      END SUB
 
      SUB PRETRE
'
'  This program determines the order in which objects appear on
'  the dendogram.
'
      DIM KLUS(100, 2)
      DIM NIN1(100)
      DIM LST(100)
      DIM JHOLD(100)
 
      OPEN "CLUM.DAT" FOR INPUT AS #3
      OPEN "SORTED.DAT" FOR OUTPUT AS #2
      INPUT #3, TITLE$
      PRINT #2, TITLE$
      INPUT #3, NN
      N = NN - 1
      PRINT #2, NN
      INPUT #3, CRIT$
      PRINT #2, CRIT$
 
      FOR I = 1 TO N
         INPUT #3, KLUS(I, 1), KLUS(I, 2), XX
         PRINT #2, USING " ### ###"; KLUS(I, 1), KLUS(I, 2)
      NEXT I
 
      FOR I = 1 TO NN
         LST(I) = I
         NIN1(I) = 1
      NEXT I
 
      FOR II = 1 TO N
 
' name tabs
 
         I = KLUS(II, 1)
         J = KLUS(II, 2)
         NI = NIN1(I)
         NJ = NIN1(J)
         L = 1
2015     IF (LST(L) = I) THEN 2020
         L = L + 1
         IF (L <= NN) THEN 2015
2020     ICOL = L
         IN = ICOL + NI
         INEND = IN + NJ - 1
         L = L + 1
2030     IF (LST(L) = J) THEN 2040
         L = L + 1
         IF (L <= NN) THEN 2030
2040     JCOL = L
         JPRE = JCOL - 1
         JEND = JCOL + NJ - 1
         NHOLD = 1
'
' remove J vector and store in HOLD
'
         FOR M = JCOL TO JEND
            JHOLD(NHOLD) = LST(M)
            NHOLD = NHOLD + 1
         NEXT M
'
' shift
'
         MSH = JEND
2055     IF (MSH = INEND) THEN 2060
         KSH = MSH - NJ
         LST(MSH) = LST(KSH)
         MSH = MSH - 1
         GOTO 2055
'
' insert hold vector
'
2060     NHOLD = 1
         FOR M = IN TO INEND
            LST(M) = JHOLD(NHOLD)
            NHOLD = NHOLD + 1
         NEXT M
         NIN1(I) = NI + NJ
      NEXT II
 
      NLINES = INT(NN / 20!) + 1
      INDX = 0
      FOR I = 1 TO NLINES
      FOR J = 1 TO 20
      INDX = INDX + 1
      IF (INDX <= NN) THEN PRINT #2, USING " ###"; LST(INDX);
      NEXT J
      PRINT #2, ""
      NEXT I
      CLOSE 2
      CLOSE 3
      END SUB
 
      SUB RMAKE (X$, LIMIT, ISTART)
'
' This subroutine prints a line of the dendogram
'
      X$ = ""
      FOR I = ISTART TO LIMIT
         X$ = X$ + ROW$(I)
      NEXT I
      END SUB
 
      SUB SCAN
'
' This subroutine looks for the maximum similarity.  It finds
' element (K, L), where K and L are the most similar objects or
' groups.
'
      N = 1
      RX = -10000!
      MN = M - 1
      FOR I = 1 TO M
         N = N + 1
         FOR J = N TO M
            IF ((RX - X(I, J)) < 0!) THEN GOTO 5
            IF ((RX - X(I, J)) = 0!) THEN GOTO 5
            IF ((RX - X(I, J)) > 0!) THEN GOTO 10
5        K = I
            L = J
            RX = X(I, J)
10     NEXT J
      NEXT I
      END SUB
 
      SUB TREE
'
' This subroutine prints a dendogram
'
      DIM KLUS(100, 2)
      DIM LST(100)
      DIM KOL(100)
 
      OPEN "SORTED.DAT" FOR INPUT AS #3
 
      STAR$ = "*"
      BLNK$ = " "
 
      INPUT #3, TITLE$
      PRINT #7, CHR$(12)
      PRINT #7, ""
      PRINT #7, TITLE$
 
      INPUT #3, NN
      INPUT #3, CRIT$
      N = NN - 1
 
      FOR I = 1 TO N
         INPUT #3, KLUS(I, 1), KLUS(I, 2)
      NEXT I
 
      FOR I = 1 TO NN
         INPUT #3, LST(I)
      NEXT I
'
' **SCALING**  The ISCL parameter controls the width of the dendogram
' by setting spacing between adjacent "stems".  ISCL depends on
' the number of cases and printer capabilities.  The preset
' values are for a 132-column printer.  Some printers permit longer
' lines, for instance, with compressed type in landscape mode, an HP
' laserprinter can print 175 columns, and ISCL values can be reset to
'
'        ISCL         NN (number of cases)
'        -----        --------------------
'          4          2 to 43
'          3          44 to 58
'          2          59 to 87
'          1          88 or higher
'
' If you have an HP or equivalent printer and want to use these
' values, remove the comment (') indicator from the following line
       IOP = 1
 
      IF (IOP = 1) THEN
        IF (NN <= 43) THEN ISCL = 4
        IF (NN > 44 AND NN <= 58) THEN ISCL = 3
        IF (NN > 59 AND NN <= 87) THEN ISCL = 2
        IF (NN > 88) THEN ISCL = 1
      ELSE
        IF (NN <= 32) THEN ISCL = 4
        IF (NN > 32 AND NN <= 43) THEN ISCL = 3
        IF (NN > 43 AND NN <= 64) THEN ISCL = 2
        IF (NN > 64) THEN ISCL = 1
      END IF
 
      LIMIT = (ISCL * NN) + 5
'
' ID's
'
      FOR II = 1 TO 3
         I = 4 - II
         FOR JK = 1 TO LIMIT
            ROW$(JK) = BLNK$
         NEXT JK
         M = 0
         FOR J = 1 TO NN
            L = LST(J)
            CALL CONV(L, I, A$)
            M = M + ISCL
            ROW$(M) = A$
         NEXT J
      X$ = "     "
      FOR KK = 1 TO LIMIT
      X$ = X$ + ROW$(KK)
      NEXT KK
      PRINT #7, X$
      NEXT II
'
' Print tree diagram
'
      ISPACE = 5
 
      FOR I = 1 TO NN
         ISPACE = ISPACE + ISCL
         LL = LST(I)
         KOL(LL) = ISPACE
      NEXT I
 
      FOR I = 1 TO 5
         ROW$(I) = BLNK$
      NEXT I
 
      L = 5
      JSCL = ISCL - 1
 
      FOR I = 1 TO NN
         IF (JSCL = 0) THEN 1226
         FOR J = 1 TO JSCL
            L = L + 1
            ROW$(L) = BLNK$
         NEXT J
1226     L = L + 1
         ROW$(L) = STAR$
      NEXT I
 
      LIMIT = (ISCL * NN) + 5
 
      FOR IT = 1 TO N
         CALL RMAKE(X$, LIMIT, 1)
         PRINT #7, X$
         MERGEI = KLUS(IT, 1)
         MERGEJ = KLUS(IT, 2)
         II = KOL(MERGEI)
         JJ = KOL(MERGEJ)
         FOR L = II TO JJ
            ROW$(L) = STAR$
         NEXT L
         PRINT #7, USING "#### "; IT;
         CALL RMAKE(X$, LIMIT, 6)
         PRINT #7, X$
 
         FOR L = II TO JJ
            ROW$(L) = BLNK$
         NEXT L
 
         RI = II
         RJ = JJ
         RKOL = (.5 * (JJ - II) + II)
         KOL(MERGEI) = INT(RKOL)
         KI = KOL(MERGEI)
         ROW$(KI) = STAR$
      NEXT IT
 
      PRINT "DENDOGRAPH PRINTED TO FILE <OUTPUT.TXT>"
      PRINT " "
      CLOSE 3
      CLOSE 7
      END SUB
 



